/**
 * @fileoverview
 *
 * ASDFetcher is a class for periodically downloading activity
 * streams from an arbitrary number or URLs. It's using the same interface
 * as the ResourceFetcher
 * @see ResourceFetcher 
 *
 * An example:
 *
 * var asd = new ASDFetcher();
 * asd.addCallback(onStreamUpdated);
 * asd.addErrback(onStreamError);
 * asd.addUrl(streamUrl1)
 * asd.addUrl(streamUrl2)
 * asd.addUrl(streamUrl3)
 *
 * When any of the streams are downloaded the onStreamUpdated function will
 * be called with two arguments: the url and the ASD stream data as
 * js objects.
 *
 * It's possible to add more than one success/error callback. All will be
 * called, but there are no guarantees they will be called in the same order
 * they were added!
 *
 * It's also possible to add success/error callbacks for each url. This can
 * be done either in the call to addUrl:
 *
 * asd.addUrl(streamUrl4, onUrl4Success, onUrl4Fail);
 *
 * or by providing a url argument to the addCallback/addErrback call:
 *
 * asd.addCallback(streamUrl3)
 *
 */
function ASDFetcher()
{
    var self = this;

    // Make sure our rsfetcher dependency is met
    if (! window.ResourceFetcher)
    {
        throw "ResourceFetcher class could not be found!\n" +
              "remember to add a script tag for resourcefetcher.js " +
              "to the html document";
    }

    // Make sure our asparser dependency is met
    if (! window.ASDParser)
    {
        throw "ActivityStreamParser class could not be found!\n" +
              "remember to add a script tag for asdparser.js " +
              "to the html document";
    }

    this._fetcher = new ResourceFetcher();
    this._parser = new ASDParser();
    this._fetcher.parser = function(txt, doc)
    {
        return self._parser.parseDocument(doc);
    }

    this.addUrl = function(url, callback, errback) { self._fetcher.addUrl(url, callback, errback); }
    this.removeUrl = function(url) { self._fetcher.removeUrl(url); }
    this.addCallback = function(cb, url) { self._fetcher.addCallback(cb, url); }
    this.addErrback = function(cb, url) { self._fetcher.addErrback(cb, url); }
    this.refresh = function(url) { self._fetcher.refresh(url); }
    this.getStatus = function(url) { return self._fetcher.getStatus(url); }
    this.toString = function() { return "[ASDFetcher instance]"; }
}
