/**
 * @fileoverview Arms Date.prototype with the following methods:
 *
 * Date.humanReadable() - returns String like: "14:40pm April 1st, 2009".
 * Date.oclock() - returns String showing time like: "12:00pm".
 * Date.getDayName()
 * Date.getMonthName()
 * Date.elapsedBy([unit = 'second'[, relationDate = now]]) - returns Number of
 *  units that separate this date and relationDate; signum(returnValue) 
 *  == signum(relationDate - this).
 * Date.happens([relationDate = now]) - returns String saying what's this date
 * in relation do the argument, e.g. 'in 2 weeks', 'last Monday at 2:00am'.
 *
 * @version 0.5
 */
(function() {

    /*
     * comprizes of methods calculating the difference between dates;
     * methods name reflects the result's unit. All the methods
     * take two Date arguments - a start Date (from when to count), 
     * and an end Date. An optional third argument defines 'toFixed'
     * precision and defaults to 0. Returns Number.
     */
    var oTimeApart = {
        /*
         * returns how many full (1000ms) seconds separate the two dates;
         */
        second : function(from, till) {
            return  +(( (till.getTime() - from.getTime())/1000 ).toFixed(arguments[2]));
        },
        /*
         * returns how many full (60sec) minutes separate the two dates;
         */
        minute : function(from, till) {
            return  +(( (till.getTime() - from.getTime())/(60*1000) ).toFixed(arguments[2]));
        },
        /*
         * returns how many full (60min) hours separate the two dates;
         */
        hour : function(from, till) {
            return  +(( (till.getTime() - from.getTime())/(60*60*1000) ).toFixed(arguments[2]));
        },
        /*
         * returns how many full (24h) days separate the two dates;
         */
        day : function(from, till) {
            return  +(( (till.getTime() - from.getTime())/(24*60*60*1000) ).toFixed(arguments[2]));
        },
        /*
         * returns how many calendar (not 24h) days separate the two dates,
         * e.g. it returnes 1 for ( 2009-10-13 10:00pm, 2009-10-14 1.00am );
         */
        calendarDay : function(from, till) {
            // 7 digits after decimal point is enough to preserve miliseconds
            var iHours = this.hour(from, till, 7);
            var iDays = iHours/24 >> 0;
            if (iHours > 0) {
                if ( till.getHours() < iHours - 24*iDays )
                    ++iDays;
            } else if (iHours < 0) {
                if ( 23 - till.getHours() < -(iHours - 24*iDays) )
                    --iDays;
            }
            return iDays;
        },
        /*
         * returns how many full (7 day) weeks separate the two dates;
         */
        week : function(from, till){
            return +(( (till.getTime() - from.getTime())/(7*24*60*60*1000) ).toFixed(arguments[2]));
        },
        /*
         * returns how many calendar (not 7 day) weeks separate 
         * the two dates (a week starts with Sunday), e.g. it 
         * returns 1 for ( Saturday the 11th, Sunday the 12th );
         * The difference must be visible in calendar days.
         */
        calendarWeek : function(from, till) {
            var iDays = this.calendarDay(from, till);
            var iWeeks = iDays/7 >> 0;
            if (iDays > 0) {
                if ( till.getDay() < (iDays - 7*iWeeks) )
                    ++iWeeks;
            } else if (iDays < 0) {
                if ( 6 - till.getDay() < -(iDays - 7*iWeeks) )
                    --iWeeks;
            }
            return iWeeks;
        },
        /*
         * returns how many calendar (not ~30 day) months separate the two dates;
         * e.g. it returns 1 for ( April 30th, May 1st )
         */
        calendarMonth : function(from, till) {
            return 12*this.calendarYear(from, till) + (till.getMonth() - from.getMonth())
        },
        /*
         * returns how many full (365.2425 day) years separate the two dates;
         */
        year : function(from, till) {
            return +(( (till.getTime() - from.getTime())/(365.2425*24*60*60*1000) ).toFixed(arguments[2]));
        },
        /*
         * returns how many calendar (not 365 day) years separate the two dates,
         * e.g. it returns 1 for ( 1410-12-31, 1411-01-01 )
         */
        calendarYear : function(from, till) {
            return till.getFullYear() - from.getFullYear();
        }
    };
    
    /*
     * returns String like: 14:40pm April 1st, 2009
     */
    Date.prototype.humanReadable = function() {
        var date = this.getDate();
        var postfix = 'th';
        if(date<4 && date>20)
        {
            postfix = ['th','st','nd', 'rd'][date%10] || 'th'
        }
        return this.oclock() +' '+this.getMonthName()+' '                           // '12:00pm April '
            +this.getDate()+(postfix)     // '27th'
            +', '+this.getFullYear();                                               // ', 1984'
    };
    /*
     * returns String showing time like: 12:00pm
     */
    Date.prototype.oclock = function() {
        var hour = this.getHours();
        var min = this.getMinutes();
        return (hour%12||12) + ':' + (min < 10 ? '0' : '') + min + ( hour<12 ? _('am') : _('pm') );
    };
    /*
     *
     */
    Date.prototype.getDayName = function() {
        return [_('Sunday'),_('Monday'),_('Tuesday'),_('Wednesday'),_('Thursday'),_('Friday'),_('Saturday')][this.getDay()];
    };
    /*
     *
     */
    Date.prototype.getMonthName = function() {
        return [_('January'),_('February'),_('March'),_('April'),_('May'),_('June'),_('July'),_('August'),_('September'),_('October'),_('November'),_('December')][this.getMonth()];
    };
        
        
    /*
     * Returns the result of oTimeApart object method indicated by unit (default is 'second'),
     * called with this as from Date and now (that defaults to new Date) as till Date;
     */
    Date.prototype.elapsedBy = function(unit, now) {
        now = now || new Date();
        unit = unit || 'second';
        
        if (!(unit in oTimeApart) || !(now instanceof Date))
            throw new Error('Date.prototype.elapsedBy: invalid argument passed')
        
        return oTimeApart[unit](this, now);
    };
    
	
	
	/*
	 * Returns the formatted date in dd.mm.yyyy hh:mm formate as needed in some display
     * @param {Dateobject} time
	 * @return {string} formated date and time 
     */
	Date.prototype.toString = function ()                     
	{
		return isNaN(this)?'Nan':('0'+this.getDate()).slice(-2)+'.'+('0'+(1+this.getMonth())).slice(-2)+'.'+this.getFullYear()+' '+('0'+this.getHours()).slice(-2)+':'+('0'+this.getMinutes()).slice(-2);
	} 
 
	
    /*
     * Returns a human readable distance beetween dates like:
     * 'this week', 'last month', '2 years ago', 'in 3 days'
     * @param {Number} apart how many periods; < 0 for future
     * @param {String} period like 'year', 'day'
     */
    function sHowMany(apart, period) {
        // These comments are all possible strings resulting from this function, its purpose is to be caught by the Translation Manager
        // _('this hour'); _('this minute'); _('this second'); _('this year'); _('this month'); _('this week');
        // _('in the next hour'); _('in the next minute'); _('in the next second'); _('in the next year'); _('in the next month'); _('in the next week'); 
        // _('in the last hour'); _('in the last minute'); _('in the last second'); _('in the last year'); _('in the last month'); _('in the last week'); 
        // _('in {number} hours'); _('in {number} minutes'); _('in {number} seconds'); _('in {number} years'); _('in {number} months'); _('in {number} weeks');
        // _('{number} hours ago'); _('{number} minutes ago'); _('{number} seconds ago'); _('{number} years ago'); _('{number} months ago'); _('{number} weeks ago');
        var s;

        if (apart == 0) {
            s = 'this ' + period;
        } 
        else if (apart == 1 || apart == -1) {
            s = ((!~apart)?('in the next '+period):('in the last '+period));
        } 
        else {    //
            if(apart < 0)
            {
                s = "in {number} " + period + "s";
            } else {
                s = "{number} " + period + "s ago";
            }
        }
        
        s = _(s, {"number": ((apart < 0)?-1:1)*apart});

        return s;
    };
    
    /*
     * Returns a human readable string saying what's the date in relation do 
     * the Date argument (that defaults to now).
     * It shows:
     * 1. if 1 - 7 calendar days separate the two dates:
     *  - yesterday/tommorrow if applicable or
     *  - last Saturday at 10.01am if it's less then a week
     * 2. if it's less than a day
     * 'right now!' if it's just happened.
     * seconds if less than a minute spearate the two dates.
     * minutes if it's less than an hour; otherwise:
     * hours
     * 3. if it's more than a week:
     * years if it's at least one calendarYear and more than enough (6) months
     * months if at least 1
     * weeks
     */
    Date.prototype.happens = function(now) {
        options = {
            enoughMonths : 6
        }
        
        now = (now instanceof Date)?now:(new Date);

        var daysApart = oTimeApart['day'](this, now);
        
        if (daysApart != 0 && daysApart > -7 && daysApart < 7) {
            // _('tomorrow'); _('yesterday');
            var s;
            if( daysApart == 1 ) {
                s = _("yesterday at {thisTime}", {"thisTime": this.oclock()});
            } else if( daysApart == -1 ) {
                s = _("tomorrow at {thisTime}", {"thisTime": this.oclock()});
            }
            else {
                // _('next Sunday at {thisTime}'); _('next Monday at {thisTime}'); _('next Tuesday at {thisTime}'); _('next Wednesday at {thisTime}'); _('next Thursday at {thisTime}'); _('next Friday at {thisTime}'); _('next Saturday at {thisTime}'); 
                // _('last Sunday at {thisTime}'); _('last Monday at {thisTime}'); _('last Tuesday at {thisTime}'); _('last Wednesday at {thisTime}'); _('last Thursday at {thisTime}'); _('last Friday at {thisTime}'); _('last Saturday at {thisTime}'); 
                s = ((daysApart < 0)?'next ':'last ') + this.getDayName() + " at {thisTime}";
                s = _(s, {"thisTime" : this.oclock()});
            } 
            return s;
        } 
        else if (daysApart == 0) {
            // these might be useful in the future:
            // _('hour'); _('minute'); _('second');
            var periods = ['hour','minute','second'];
            for (var i = 0, period; period = periods[i]; ++i) {
                if (apart = oTimeApart[period](this, now)) {
                    return sHowMany(apart, period);
                }
            }
            return _('right now!');
        } 
        else {
            // _('year'); _('month'); _('week'); 
            var monthsApart = oTimeApart['calendarMonth'](this, now);
            var yearsApart = oTimeApart['calendarYear'](this, now);
            
            if (yearsApart && Math.abs(monthsApart) > options.enoughMonths) {
                return sHowMany(yearsApart, 'year');
            } 
            else if (monthsApart) {
                return sHowMany(monthsApart, 'month');
            } 
            else {
                var weeksApart = oTimeApart['calendarWeek'](this, now);
                return sHowMany(weeksApart, 'week');
            }
        }
    
    };

    /**
     * Parse a date/time in rfc 3339 format into a js date object
     * This code is taken from the widgetize atom parser!
     */
    String.prototype.parseRfcDate = function() {
        var rfc3339 = /(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2})(\.\d+)?(Z|[+-]\d{2}:\d{2})/i;
        var res, hoff, moff; 
        if (res = rfc3339.exec(this)) {
            if (res[8] == 'Z') {
                hoff = 0;
                moff = 0;
            } else {
                var s = res[8].split(':');
                hoff = + s[0];
                moff = + s[1];
            }
            var ms = res[7] ? res[7] * Math.pow( 10, -res[7].length ) : 0;
            return new Date( Date.UTC(res[1], res[2] - 1, res[3], res[4] - hoff, res[5] - moff, res[6], ms ));
        } else {
            return null;    
        }
    }
})();

