/**
 * Copyright (c) 2009, Opera Software ASA
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Opera Software ASA nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY OPERA SOFTWARE ASA AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL OPERA SOFTWARE ASA AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */


/**
 * @fileoverview
 * Yusef's ACL plugin.
 *
 *
 * @author Mathieu 'p01' Henri, Opera Software ASA (p01@opera.com)
 * @author Antonio Afonso, Opera Software ASA (antonio.afonso@opera.com)
 * @version 1.0
 */

if( opera.io.webserver.userName!='' )
{
    Yusef.addPlugin
    (
        'acl',
        /**
         *  ACL has no public constructor.
         *
         *  @class ACL
         *
         *  <p>The ACL plugin is available via Yusef.plugins.acl and provides Unite Applications with a customizable access control list.</p>
         *
         *      <p style="text-align:center">&#x2042;</p>
         *
         *  <h1>Access types:</h1>
         *  <p>There are 3 access types:</p>
         *
         *  <h2>public</h2>
         *  <p>Anyone can access such resource.</p>
         *
         *  <h2>limited</h2>
         *  <p>A password is required to access such resource.</p>
         *  <p>Access to such resource is gained via the Unite Action <code>Yusef_plugins_acl.unlock</code>.</p>
         *  <p>The password can be modified using the Unite Action <code>Yusef_plugins_acl.change</code>. Changing the password of a resource will revoke access to the visitors who entered the previous password.</p>
         *  <p><strong>Notice:</strong> The password can only contain Latin characters, numbers and underscore. It is considered insecure, but accepted for convenience reasons, if it is less than 4 characters long.</p>
         *
         *  <h2>private</h2>
         *  <p>Only the owner of the Unite Application can access such resource.</p>
         *
         *      <p style="text-align:center">&#x2042;</p>
         *
         *  <h1>Access nodes:</h1>
         *  <p>Access nodes are points ( URLs ) in the Unite application which have an access type bound to them.</p>
         *  <p>The access node can be strict or not. Strict access nodes can not be changed. Other access nodes can be modified using the Unite Action <code>Yusef_plugins_acl.change</code>.</p>
         *  <p>Yusef.addSectionListener() is extended to create an access node for the section if the developer wants.</p>
         *  <p><strong>Notice:</strong> For privacy reasons, the ACL plugin sets the root of the Unite Application to limited access by default. The developer or administrator of the Unite Application can modify this at will.</p>
         *
         *  <p>The properties of an Access Node are:</p>
         *
         *  <dl>
         *
         *      <dt>path</path>
         *      <dd>A String representing the path of the access node.</dd>
         *
         *      <dt>type</path>
         *      <dd>A String representing the type of access of the access node.</dd>
         *
         *      <dt>strict</path>
         *      <dd>A Boolean representing whether this access node is strict or not.</dd>
         *
         *  </dl>
         *
         *      <p style="text-align:center">&#x2042;</p>
         *
         *  <h1>Unite Actions:</h1>
         *
         *  <h2>Yusef_plugins_acl.unlock</h2>
         *  <p>The Unite Action <code>Yusef_plugins_acl.unlock</code> is registered to handle the need for a visitor to gain access to a node set to Limited access.</p>
         *  <p>The arguments of this Unite Action are:</p>
         *
         *  <dl>
         *
         *      <dt>unite-action</dt>
         *      <dd><p>The name of the Unite Action: <code>Yusef_plugins_acl.unlock</code></p></dd>
         *
         *      <dt>unite-nonce</dt>
         *      <dd><p>The cryptographic nonce provided by Yusef, on <code>connection.session.nonce</code>.</p></dd>
         *
         *      <dt>password</dt>
         *      <dd><p>The password to unlock the access to a <code>limited</code> resource.</p></dd>
         *
         *  </dl>
         *
         *  <h2>Yusef_plugins_acl.change</h2>
         *  <p>The Unite Action <code>Yusef_plugins_acl.change</code> is registered to handle the need for the owner to change the type of an access node.
         *  <p>The arguments of this Unite Action are:</p>
         *  <dl>
         *
         *      <dt>unite-action</dt>
         *      <dd><p>The name of the Unite Action: <code>Yusef_plugins_acl.change</code></p></dd>
         *
         *      <dt>unite-nonce</dt>
         *      <dd><p>The cryptographic nonce provided by Yusef, on <code>connection.session.nonce</code>.</p></dd>
         *
         *      <dt>accessNode</dt>
         *      <dd><p>The path of the access node the change will effect.</p></dd>
         *
         *      <dt>type</dt>
         *      <dd><p>The type of access applied to access node. <code>public</code>, <code>limited</code> or <code>private</code>.</p></dd>
         *
         *      <dt>password</dt>
         *      <dd><p>The password for the <code>limited</code> access type.</p></dd>
         *
         *  </dl>
         *
         *      <p style="text-align:center">&#x2042;</p>
         *
         *  <h1>Extensions made to Yusef</h1>
         *  <p>The ACL plugin extends the following methods of Yusef in order to inject the access control checks.</p>
         *
         *  <h2>Yusef.registerUniteAction</h2>
         *  <p>This method now takes an optionnal third argument with the property <code>acl.level</code> which specifies the required level to perform the Unite Action.</p>
         *
         *  <pre class="code">
         *  //  Register the 'get-file-information' and allow it for everyone.
         *
         *  var accessTypes = Yusef.plugins.acl.getAccessTypes();
         *  Yusef.registerUniteAction
         *  (
         *      'get-file-information',
         *      myFileDownloader,
         *      {
         *          'acl':
         *          {
         *              'level':    accessTypes.public.level
         *          }
         *      }
         *  );
         *  </pre>
         *
         *  <pre class="code">
         *  //  Register the 'upload-file' and only allow it for the adminstrator of the Unite Application.
         *
         *  var accessTypes = Yusef.plugins.acl.getAccessTypes();
         *  Yusef.registerUniteAction
         *  (
         *      'upload-file',
         *      myFileUploader,
         *      {
         *          'acl':
         *          {
         *              'level':    accessTypes.private.level
         *          }
         *      }
         *  );
         *  </pre>
         *
         *  <h2>Yusef.addSectionListener</h2>
         *  <p>This method now takes an optionnal third argument with the properties <code>acl.type</code> and <code>acl.strict</code> which respectively specify the access type of the section and whether the access node associated to the section is strictly of that type or can be changed using the Unite Action 'Yusef_plugins_acl.change'.</p>
         *  <p><strong>Notice:</strong> Adding a section listener with an type sets an access node for the URL of the section.</p>
         *
         *  <pre class="code">
         *  //  Adds a section listener for 'admin_area' whose access will be strictly private.
         *
         *  Yusef.addSectionListener
         *  (
         *      'admin_area',
         *      myAdminAreaSectionListener,
         *      {
         *          'acl':
         *          {
         *              'type':    'private',
         *              'strict':   true
         *          }
         *      }
         *  );
         *  </pre>
         *
         *  <pre class="code">
         *  //  Adds a section listener for 'stuff' whose access will be limited by default but could be changed via the Unite Action 'Yusef_plugins_acl.change'.
         *
         *  Yusef.addSectionListener
         *  (
         *      'stuff',
         *      myAdminAreaSectionListener,
         *      {
         *          'acl':
         *          {
         *              'type':    'limited'
         *          }
         *      }
         *  );
         *  </pre>
         *
         */
        function ACL()
        {
            var that            = this;
            var undefined;

            /**
             *  @private
             *  @constant
             */
            var _config = {};
            _config.pluginName          = 'YUSEF\'s ACL PLUGIN';

            /**
             *  The name of the UniteAction to unlock access to a node set to Limited access is 'Yusef_plugins_acl.unlock'.
             *  @constant
             */
            _config.actionName_unlock   = 'Yusef_plugins_acl.unlock';
            /**
             *  The name of the UniteAction to change the type of access of a node is 'Yusef_plugins_acl.change'.
             *  @constant
             */
            _config.actionName_change   = 'Yusef_plugins_acl.change';
            _config.sessionExpire       = 1000*60*60*24*30; // 30 days
            _config.sessionSoftMax      = 8;
            _config.sessionPurgeDelay   = 1000*60*60*24;    // 1 day
            _config.passwordMinLength   = 4;
            _config.passwordMaxLength   = 64;

            _config.sharedPublicHtmlFolder


            /**
             *  @private
             */
            var _accessTypes    = {};
            /**
             *  @private
             */
            var _sessions       = {};
            /**
             *  @private
             */
            var _accessNodes    = {};


            //  --
                PSO.pub( 'purgeAndSaveSessions', _enumerateObject( _sessions ,'_session before -- FALLBACK' ) );
            var savedInfo = widget.preferenceForKey( _config.pluginName+'_sessions' )||'';
            if( savedInfo!=='' )
            {
                _sessions = JSON.parse( savedInfo );
                PSO.pub( 'purgeAndSaveSessions', _enumerateObject( _sessions ,'_session before -- RESTORE' ) );
                _purgeAndSaveSessions();
            }


            //  --
            var savedInfo = widget.preferenceForKey( _config.pluginName+'_accessNodes' )||'';
            if( savedInfo!=='' )
            {
                _accessNodes = JSON.parse( savedInfo );
            }


            /**
             *  Generate a password made of alphanumeric characters.
             *
             *  @private
             *  @param  {Number}    length  The length of the desired password. Default length is 8 characters.
             *  @return {String}
             */
            function _generatePassword( len )
            {
                var len = len||8;
                return (new Array(len).join("0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_")).split('').sort(function(){return Math.random()-.5}).slice(0,len).join('');
            }


            /**
             *  Create an accessType object.
             *
             *  @private
             *  @param  {Number}    level   The level
             *  @param  {Number}    level   The level
             *  @param  {Number}    level   The level
             */
            function _createAccessType( level, uid, type, hasAccess, accessDenied, accessGranted, meta )
            {
                _accessTypes[ uid ] =
                {
                    'level'         : level,
                    'uid'           : uid,
                    'type'          : type,
                    'hasAccess'     : hasAccess,
                    'accessDenied'  : accessDenied,
                    'accessGranted' : accessGranted,
                    'meta'          : meta,
                    'locked'        : level===Infinity
                }


                //  --
                var savedMeta = widget.preferenceForKey( _config.pluginName+'_accesTypes['+type+'].meta' )||'';
                if( savedMeta!=='' )
                {
                    var savedMeta = JSON.parse( savedMeta );
                    for( var k in savedMeta )
                    {
                        if(k=="password" || k=="passwordIsSecure" || k=="magic" )
                          _accessTypes[ type ].meta[ k ] = savedMeta[k];
                    }
                }
                widget.setPreferenceForKey( JSON.stringify( _accessTypes[ type ].meta ), _config.pluginName+'_accesTypes['+type+'].meta' );
                widget.setPreferenceForKey( JSON.stringify( _accessNodes ), _config.pluginName+'_accessNodes' );
            }


            //  limited
            _createAccessType
            (
                2,
                'limited',
                'limited',
                function _accessType_limited_hasAccess( connection )
                {
                    PSO.pub( '_accessTypes', _enumerateObject( this, 'this -- hasAccess_limited' ) );

                    if( connection.sessionId )
                    {
                        PSO.pub
                        (
                            'Yusef.plugins.acl infos',
                            'hasAccess_limited -> connection.sessionId  = '+ connection.sessionId,
                            '_sessions[ connection.sessionId ] = '+ (''+_sessions[ connection.sessionId ] )
                        );
                        var acl = _sessions[ connection.sessionId ];
                        if( acl!==undefined )
                        {
                            PSO.pub
                            (
                                'Yusef.plugins.acl infos',
                                _enumerateObject( acl, '_sessions[ '+ connection.sessionId +' ]' )
                            );

                            if( acl.passwords!==undefined )
                            {
                                if( acl.passwords[ this.meta.password ]===true )
                                {
                                    return true;
                                }
                            }
                        }
                    }

                    return false;
                },
                function _accessType_limited_accessDenied( connection )
                {
                    var obj =
                    {
                        'HTMLContent'       : _('Sorry, you need a password to access this resource.'),
                        'unite-actions'  :
                        [
                            {
                                id          :_config.actionName_unlock,
                                headline    :_('Contact {userName} and enter the password.', {userName: opera.io.webserver.userName.HTMLEntities()}),
                                fields      :
                                [
                                    {name:'password','label':_('The password'),'type':'password'}
                                ]
                            }
                        ]
                    };
                    return obj;
                },
                null,
                {
                    'label'             :__('Password protected'),
                    'summary'           :__('Visitors must know the password'),
                    'description'       :__('Only for friends with the password.')/*,
                    'password'          :_generatePassword( _config.passwordMinLength ),
                    'passwordIsSecure'  : true
                    */
                }
            );


            /**
             *  @private
             *  Sets and save the password of the limited accessType and the associated meta data such as passwordIsSecure and magic
             */
            function _accessType_limited_setPassword( password )
            {
                var meta = _accessTypes.limited.meta;
                var changing = meta.magic==undefined || password!==meta.password;
                meta.password          = password;
                meta.passwordIsSecure  = password.length>=_config.passwordMinLength;
                if( changing )
                {
                    meta.magic             = _hash( new Date().getTime()+Math.random() );
                }
//              meta.magic             = _hash( password+'\n'+widget.identifier+'\n'+opera.io.webserver.userName );
                widget.setPreferenceForKey( JSON.stringify( meta ), _config.pluginName+'_accesTypes[limited].meta' );
            }
            
            //  generate a default password
            _accessType_limited_setPassword( _accessTypes.limited.meta.password||_generatePassword( _config.passwordMinLength ) );


            //  public
            _createAccessType
            (
                1,
                'public',
                'public',
                function _accessType_public_hasAccess( connection )
                {
                    return true;
                },
                null,
                null,
                {
                    'label'         :__('Public'),
                    // FIX: since this can't be properly translated we're dropping this string in the meanwhile.
                    //'summary'       :_('Anyone may visit your {serviceName}', {serviceName: opera.io.webserver.currentServiceName}),
                    'summary'       :__('Anyone may visit your {serviceName}', {serviceName: opera.io.webserver.currentServiceName}),
                    'description'   :__('Open for everyone.')
                }
            );


            //  private
            _createAccessType
            (
                Infinity,
                'private',
                'private',
                function _accessType_private_hasAccess( connection )
                {
                    return connection.isOwner;
                },
                function _accessType_private_accessDenied( connection )
                {
                    var obj =
                    {
                        'HTMLContent'   : _('Sorry, this resource can only be accessed from the Opera Unite running this application. ')
                    };
                    if( connection.isLocal )
                    {
                        obj.content +=_(' <a href="#">Use the Unite panel to start the application.</a>');
                    }
                    return obj;
                },
                function _accessType_private_accessGranted( connection )
                {
                    var fields = [];
                    fields.push
                    (
                        {
                            'name':'accessNode',
                            'type':'hidden',
                            'value':connection.request.accessNode.uid
                        }
                    )
                    for( var id in _accessTypes )
                    {
                        var field = {};
                        field.name  = 'type';
                        field.label = id +' -- '+ _accessTypes[id].meta.summary
                        field.type  = 'radio';
                        field.value = id;
                        if( connection.request.accessType==id )
                        {
                            field.checked   = 'checked';
                        }
                        fields.push( field );

                        if( _accessTypes[id].meta.password )
                        {
                            fields.push( {'name':'password','label':_('The password <em>{password}</em>', {password: _accessTypes[id].meta.password}),'type':'password','value':_accessTypes[id].meta.password} );
                        }

                    }
                    var obj =
                    {
                        'unite-actions'  :
                        [
                            {
                                id          :_config.actionName_change,
                                headline    : _('Change the access type of <em>{uid}</em>', {uid: connection.request.accessNode.uid}),
                                fields      : fields
                            }
                        ]
                    };
                    return obj;
                },
                {
                    'summary'       :_('Only you can access these files.'),
                    'description'   :_('Only for you.')
                }
            );


            /**
             *  Disable an access type.
             *  <p><strong>Notice:</strong> It is not possible to disable an access type that is in use.</p>
             *
             *  @public
             *  @param  {String}    type    The type of access to disable.
             */
            this.disableAccessType = function( type )
            {
                var deleted = false;
                if( _isValidAccessType( type ) )
                {
                    if( _accessTypes[type].locked===false )
                    {
                        deleted = true;
                        delete _accessTypes[type];
                    }
                    else
                    {
                        throw { name:_config.pluginName, message:'ACCESS TYPE "'+ type +'" ALREADY LOCKED' };
                    }
                }
                PSO.pub( '_accessTypes', (deleted?'Did':'Could NOT')+' delete the _accessTypes['+type+']\n'+_enumerateObject( _accessTypes, '_accessTypes' ) );

            }


            /**
             *  _isValidAccessType
             *
             *  @private
             */
            function _isValidAccessType( type )
            {
                return typeof(type)==='string' &&  (type in _accessTypes);
            }


            /**
             *  Get the list of access types.
             *
             *  @public
             *  @return {Object}   A hash map of the informations about the various access types, with the uid of the access types in key.
             *
             *  <pre class="code">
             *  //  Get the password of the Limited acces type for display on a private resource.
             *
             *  var accessTypes         = Yusef.plugins.acl.getAccessTypes();
             *
             *  var password            = accessTypes.limited.meta.password;
             *  var passwordIsSecure    = accessTypes.limited.meta.passwordIsSecure;
             *  </pre>
             */
            this.getAccessTypes = function()
            {
                return _copyObject( _accessTypes );
            }


            /**
             *  Get the list of access nodes.
             *
             *  @public
             *  @return {Object}    A hash listing the access nodes, URLs, which have an access type bound to them. The URL of the nodes is the key in that hash map.
             */
            this.getAccessNodes = function()
            {
                return _copyObject( _accessNodes );
            }


            /**
             *  Get the access node a resource belong to.
             *
             *  @public
             *  @param  {WebServerConnection}   connection
             *  @param  {String}    section     Optional. The section where the resource is.
             *  @param  {String}    path     Optional. The path of the resource in the section.
             *
             *  @return {Object}    Return the accessNode object to which resource belong.
             */
            this.getAccessNode = function( connection, section, path )
            {
                var section = ''+(section||connection.request.section||'');
                var path    = ''+(path||connection.request.path||'/');

                //  check in which _accessNodes does the resource belong
                var partialPath = '/'+section+path;

                while( _accessNodes[partialPath]===undefined )
                {
                    if( partialPath.slice(-1)==='/' )
                    {
                        partialPath = partialPath.slice( 0, -1 );

                    }
                    partialPath = partialPath.slice( 0, partialPath.lastIndexOf( '/' )+1 );
                }

                return _accessNodes[partialPath];
            }


            /**
             *  Set an access node.
             *
             *  @public
             *  @param  {String}    fullpath    The full path of the access node, starting right after from the
             *  @param  {String}    type        The type of access desired for this access node.
             *  @param  {Boolean}   strict      Optional. False by default. Define if the access node is strictly of the desired type of if the access type of this node can be changed later.
             */
            this.setAccessNode = function( serviceWidePath, type, strict )
            {
                var path = serviceWidePath;

                //  correct path ?
                if( typeof(path)!=='string' ) //|| path[0]!=='/' )
                {
                    throw { name:_config.pluginName, message: 'INCORRECT PATH "'+ path +'"' };
                }

                //  valid AccessType ?
                if( !_isValidAccessType(type) )
                {
                    throw { name:_config.pluginName, message: 'INVALID ACCESS TYPE' };
                }

                //  strict accessNode ?
                if( _accessNodes[path]!==undefined && _accessNodes[path].strict===true )
                {
                    throw { name:_config.pluginName, message: 'STRICT NODE' };
                }

                _accessNodes[ path ] =
                {
                    'path'      : path,
                    'type'      : type,
                    'strict'    : strict||false
                };
                _accessTypes[ type ].locked   = true;
                
                _setAuthentication();
            }


            /**
             *  Determine whether the visitor has access to a given resource.
             *
             *  @public
             *  @param  {WebServerConnection}   connection
             *  @param  {String}    section     The name of the section the resource belong to.
             *  @param  {String}    path        The path of the resource in the section.
             *
             *  @return {Boolean}   True if the visitor has access to the resource. False otherwise.
             */
            this.hasAccess = function( connection, section, path )
            {
                var accessNode = this.getAccessNode( connection, section, path );

                if( arguments.length===1 )
                {
                    connection.request.accessNode = accessNode;
                    connection.request.accessType = accessNode.type;
                }

                //  hasAccess ?
                return connection.session.acl.hasAccessTo[ accessNode.type ]||false;
            }

            /**
             *  Return some data about the ACL, the current application & connection.
             *
             *  @public
             *  @param  {WebServerConnection}   connection
             *
             *  @return {Object}   True if the visitor has access to the resource. False otherwise.
             */
            this.getData = function( connection )
            {
                var data            = {};

                //  is there a session, with ACL information ?
                if( connection.session && connection.session.acl )
                {
                    var accessNode      = this.getAccessNode( connection );
                    data.hasAccess      = this.hasAccess( connection );
                    data.isPublic       = accessNode.type == 'public';
                    data.isPrivate      = accessNode.type == 'private';
                    data.isRestricted   = accessNode.type == 'limited';
                    data.aclTypes       = this.getAccessTypes();
                    data.acl            = connection.session.acl;
                    data.acl.magic      = _accessTypes.limited.meta.magic;
                    data.acl.type       = accessNode.type;
                    data.accessNode     = accessNode.path;
                    if( connection.isOwner && _config.sharedPublicHtmlFolder )
                    {
                        data.js         = 'static/acl/yusef.plugins.acl.js';
                    }
                    data.actions        =
                    {
                        change          : _accessTypes.private.accessGranted( connection )['unite-actions'][0].id
                    }
                }

                return data;
            }
                
            /**
             *  manageSessionAndAcl
             *  @private
             */
            function _manageSessionAndAcl( connection )
            {
                //  mock session object
                var acl         = {};
                acl.hasAccessTo = {};
                acl.passwords   = {};
                acl.level       = 0;

                //  a sessionId ?
                if( connection.sessionId )
                {
                    //  _sessions object
                    if( _sessions[ connection.sessionId ] )
                    {
                        var acl = _sessions[ connection.sessionId ];
                    }
                    else
                    {
                        _sessions[ connection.sessionId ] = acl;
                    }
                }

                acl.lastTime = new Date().getTime();

                //  initial pass to figure the immediate access level and hasAccessTo
                for( var id in _accessTypes )
                {
                    //  skip if the session already has a higher access level
                    if( _accessTypes[ id ].level<acl.level )
                    {
                        continue;
                    }

                    var hasAccess = _accessTypes[ id ].hasAccess( connection );
                    if( hasAccess===true && _accessTypes[ id ].level>acl.level )
                    {
                        acl.level = _accessTypes[ id ].level;
                    }
                    acl.hasAccessTo[ id ] = hasAccess;
                }

                //  second pass ( based on the final access level ) to override the hasAccessTo of lesser level
                for( var id in _accessTypes )
                {
                    if( _accessTypes[ id ].level<acl.level )
                    {
                        acl.hasAccessTo[ id ] = true;
                    }
                }

                //  purgeAndSaveSessions
                _purgeAndSaveSessions( connection.sessionId );

                //  copy the acl object onto connection.session.acl
                connection.session.acl = _copyObject( acl );

                PSO.pub( 'Yusef.plugins.acl infos', connection.request.uri, _enumerateObject( connection.session.acl, 'connection.session.acl' ) ) //, _enumerateObject( _sessions, '_sessions' ) );
            }


            /*
             *  purgeAndSaveSessions
             */
            function _purgeAndSaveSessions( save )
            {
                var save            = save||false,
                    sessionsCount   = 0,
                    timeLimit       = new Date().getTime()-_config.sessionExpire;

                PSO.pub( 'purgeAndSaveSessions', _enumerateObject( _sessions ,'_session before' ) );

                //  purge
                for( var id in _sessions )
                {
                    if( _sessions[id].lastTime<timeLimit )
                    {
                        _sessions[id] = null;
                        delete _sessions[id];
                        save    = true;
                    }
                    else
                    {
                        sessionsCount++;
                    }
                }

                //  save?
                if( save )
                {
                    widget.setPreferenceForKey( JSON.stringify( _sessions ), _config.pluginName+'_sessions' );
                }

                //  exceed the _config.sessionSoftMax schedule another call
                if( sessionsCount>_config.sessionSoftMax )
                {
                    setTimeout( arguments.callee.bind( that ), _config.sessionPurgeDelay );
                }

                PSO.pub( 'purgeAndSaveSessions', _enumerateObject( _sessions ,'_session after' ) );
            }


            /**
             *  Sets the authentication flag on the service in opera.io.webserver.services.
             *  @private
             */
            function _setAuthentication()
            {
                var authentication = 0;
                //  check all the access nodes
                for( var k in _accessNodes )
                {
                    if( _accessNodes[k].type==='limited' )
                    {
                        authentication = 1;
                    }
                }
                
                //  set the authentication flag
                try
                {
                    opera.io.webserver.passwordProtected = authentication;
                }
                catch(e)
                {
                }
                return false;
            }


            /*
             *  register uniteAction listener <config.actionName_unlock> to allow the visitors to unlock an accessType
             *  for everyone
             */
            function _uniteActionListener_unlock( connection )
            {
                var found       = false;
                var servicePathHash = (connection.request.bodyItems['unite-servicePathHash']||[''])[0];
                var password = connection.request.bodyItems[servicePathHash+'password'][0];
                if( password!=='' )
                {
                    for( var id in _accessTypes )
                    {
                        if( !found && _accessTypes[id].meta.password==password )
                        {
                            found   = true;
                            _sessions[ connection.sessionId ].passwords[ password ] = true;
                        }
                    }
                }

                if( !found )
                {
                    return { 'errorMessage':_('Invalid password') };
                }
            }


            /*
             *  the private method dealing with ACL changes ( only for he owner )
             *
             */
            function _uniteActionListener_change( connection )
            {
                var found       = false;
                var type        = connection.request.bodyItems['type'][0];
                var path        = connection.request.bodyItems['accessNode'][0];
                var password    = connection.request.bodyItems['password'][0];

                try
                {
                    Yusef.plugins.acl.setAccessNode( path, type );

                    if( _accessTypes[ type ].meta.password!==undefined )
                    {
                        var pattern = '^\\w{'+ parseInt( 1 /*+0*_config.passwordMinLength*/ ) +','+ parseInt(_config.passwordMaxLength) +'}$';
                        if( !password.match( new RegExp( pattern ) ) )
                        {
                            return { 'errorMessage': _('Invalid password. Permitted characters are a-z, A-Z and 0-9.') };
                        }
                        else
                        {
                            _accessType_limited_setPassword( password );
                        }
                    }
                }
                catch( e )
                {
                    PSO.pub( 'Yusef.plugins.acl', e );
                    return e;
                }

                //  save a loose copy of _accessNodes
                var _looseAccessNodes = _copyObject( _accessNodes);
                for( var k in _looseAccessNodes )
                {
                    if( _looseAccessNodes[k].strict===true )
                    {
                        _looseAccessNodes[k] = null;
                        delete _looseAccessNodes[k];
                    }
                }
                widget.setPreferenceForKey( JSON.stringify( _looseAccessNodes ), _config.pluginName+'_accessNodes' );
            }

            function _getMagicFromQueryItem( connection )
            {
                return (connection.request.queryItems._magic||[])[0];
            }

            function _isValidMagic( magic )
            {
                return magic === _accessTypes.limited.meta.magic;
            }
            
            function _unlockPasswordUsingMagic( connection, magic )
            {
                magic = magic || _getMagicFromQueryItem(connection);
                
                if( _isValidMagic(magic) )
                {
                    _sessions[ connection.sessionId ].passwords[ _accessTypes.limited.meta.password ] = true;
                    
                    return true;
                }
                
                return false;
            }
            
            /**
             *  Initialize method.
             *  Setup the API routes to
             */
            this.initialize = function()
            {
                /*
                 *  extend Yusef.registerUniteAction to inject the ACL test in the listener
                 *  options.acl.level = minimum level required to process the uniteAction
                 */
                Yusef.extend
                (
                    'registerUniteActionListener',
                    function _extendedRegisterUniteActionListener( name, listener, options )
                    {
                        var options = options||{};
                        var level   = (options.acl||{level:Infinity}).level;

                        /**
                         *  modify the listener to check the ACL
                         *  @ignore
                         */
                        var modifiedListener_extendedRegisterUniteActionListener = function( connection )
                        {
                            //  session+acl management
                            _manageSessionAndAcl( connection );

                            //  has required level ?
                            if( connection.session.acl.level >= level )
                            {
                                return listener.apply( this, arguments ) || {};
                            }

                            return false;
                        }

                        PSO.pub( 'modified Yusef.registerUniteActionListener', name, listener, modifiedListener_extendedRegisterUniteActionListener, _enumerateObject( options, 'options' ) );


                        //  proceed with the modifiedListener
                        return _extendedRegisterUniteActionListener.proceed( name, modifiedListener_extendedRegisterUniteActionListener, options );
                    }
                );


                /*
                 *  Extend Yusef.addRoute to inject the ACL test in the method.
                 */
                Yusef.extend
                (
                    'addRoute',
                    function _extentedAddRoute( pattern, method, options )
                    {
                        var options = options||{};
                        var level   = (options.acl||{level:Infinity}).level;
                        
                        /*
                         *  Modify the method to check the ACL
                         *  @ignore
                         */
                        var modifiedMethod_extentedAddRoute = function( connection, matches )
                        {
                            //  session+acl management
                            _manageSessionAndAcl( connection );
                            
                            if( _unlockPasswordUsingMagic(connection) )
                            {
                                // re-evaluate the access levels
                                _manageSessionAndAcl( connection );
                            }
                            
                            //  hasAccess ?
                            connection.hasAccess = Yusef.plugins.acl.hasAccess( connection );
                            connection.hasAccess |= connection.session.acl.level>=level;

                            //  connection.hasAccess
                            if( connection.hasAccess )
                            {
                                var output = method.apply( this, arguments ) || {};

                                for( var type in _accessTypes )
                                {
                                    if( _accessTypes[ type ].accessGranted && connection.session.acl.hasAccessTo[ type ]==true )
                                    {
                                        var accessGrantedOutput = _accessTypes[ type ].accessGranted( connection );
                                        for( var k in accessGrantedOutput )
                                        {
                                            if( !output[k] )
                                            {
                                                output[k] = accessGrantedOutput[k];
                                            }
                                        }
                                    }
                                }
                            }
                            else
                            {
                                connection.response.setResponseHeader( 'X-Unite-ACL', 'Access Denied' );
                                var output  = _accessTypes[ connection.request.accessType ].accessDenied( connection );
                            }

                            return output;
                        }         


                        //  proceed
                        return _extentedAddRoute.proceed( pattern, modifiedMethod_extentedAddRoute, options );
                    }
                );


                /*
                 *  extend Yusef.addSectionListener to inject the ACL test in the listener
                 */
                //  TODO: extend addSectionListener and addRoute seperately. The first one to take care of the _index case
                Yusef.extend
                (
                    'addSectionListener',
                    /**
                     *  Extends Yusef.addSectionListener to support ACL.
                     *  <p>The ACL plugin extends Yusf.addSectionListener to make it accept  third argument called options </p>
                     */
                    function _extendedAddSectionListener( name, listener, options )
                    {
                        

                        var path    = (name==='_index'?'/':'/'+name+'/');
                        var options = options||{};
                        var acl     = options.acl||{};
                        var type    = acl.type||false;
                        var strict  = acl.strict||false;

                        if( type )
                        {
                            //  valid AccessType ?
                            if( !_isValidAccessType(type) )
                            {
                                throw { name:_config.pluginName, message: 'INVALID ACCESS TYPE "'+ type +'"' };
                            }
                            //  known accessNode ?
                            if( _accessNodes[path]!==undefined )
                            {
                                type = _accessNodes[path].type;
                            }
                        }
                        
                        var success = _extendedAddSectionListener.proceed( name, listener, options );
                        //  success -> set accessNode
                        if( success && type )
                        {
                            that.setAccessNode( path, type, strict||false );
                            PSO.pub( 'Yusef.addSectionListener', ''+ _enumerateObject( _accessNodes, '_accessNodes' ) );
                        }
                        return success;
                    }
                );



                Yusef.registerUniteActionListener
                (
                    _config.actionName_unlock,
                    _uniteActionListener_unlock,
                    {
                        acl:
                        {
                            level:1
                        }
                    }
                );


                /*
                 *  register uniteAction listener <config.actionName_change> to allow the owner to change the accessType of a node
                 *  only for the Owner
                 */
                Yusef.registerUniteActionListener
                (
                    _config.actionName_change,
                    function( connection )
                    {
                        var value = _uniteActionListener_change( connection );
                        if( value )
                        {
                            return value;
                        }

                        //  redirect to the same URI to prevent
                        connection.response.setStatusCode( '302', 'Found' );
                        connection.response.setResponseHeader( 'Location', 'http://' + connection.request.headers.Host[0] + connection.request.uri );
                        connection.response.close();

                    },
                    {
                        acl:
                        {
                            level:Infinity
                        }
                    }
                );


                //  force '/*' to 'limited' by default
                if( _accessNodes['/']===undefined )
                {
                    this.setAccessNode( '/', 'limited', false );
                }

                //  public_html ?
                var path = this.path||'mountpoint://application/libraries/yusef/plugins/acl';
                var public_html = Yusef.mountpoints.application.resolve( path+'/public_html' );
                _config.sharedPublicHtmlFolder  = false;
                if( public_html.exists )
                {
                    Yusef.shareStaticDirectory( public_html,'acl/' );
                    _config.sharedPublicHtmlFolder  = true;
                }

                
                //  magic Route
                Yusef.addRoute
                (
                    /magic\/([^\/]+)\/?/,
                    function( connection, matches )
                    {
                        //  session+acl management
                        _manageSessionAndAcl( connection );

                        //  valid magic ?
                        //var validMagic  = matches[0]===_accessTypes.limited.meta.magic;
                        //if( validMagic )
                        if( _unlockPasswordUsingMagic(connection, matches[0]) )
                        {
                            //  check the redirection path
                            var redirectionPath = (connection.request.getItem( 'redir' )||[''])[0];
                            if( redirectionPath.match( /\/\./ ) )
                            {
                                redirectionPath = '';
                            }
                            redirectionPath = opera.io.webserver.currentServicePath+redirectionPath;

                            //  unlock the password 
                            //_sessions[ connection.sessionId ].passwords[ _accessTypes.limited.meta.password ] = true;
                            
                            //  redirect
                            connection.response.setStatusCode( '303', 'See Other' );
                            connection.response.setResponseHeader( 'Location', redirectionPath );
                            connection.response.close();
                            return false;
                        }

                        //  redirect requests with an invalid magic to the index of the service
                        connection.response.setStatusCode( '303', 'See Other' );
                        connection.response.setResponseHeader( 'Location', opera.io.webserver.currentServicePath );
                        connection.response.close();
                    },
                    {
                        'acl':{ 'level':0 }
                    }
                );


                //  APIRoute '/' -> update session, give nonce, etc...
                this.addAPIRoute
                (
                    '1.0',
                    '',
                    function( connection, matches )
                    {
                        //  session+acl management
                        _manageSessionAndAcl( connection );
                    }
                );

                //  APIRoute '/unlock/' -> update session, give nonce, etc... unlock access
                this.addAPIRoute
                (
                    '1.0',
                    /unlock/,
                    function( connection, matches )
                    {
                        //  session+acl management
                        _manageSessionAndAcl( connection );

                        if( connection.request.method!=='POST' || !connection.session.isValidNonce )
                        {
                            return;
                        }

                        //  _uniteActionListener_unlock -> an error object ?
                        if( _uniteActionListener_unlock( connection ) )
                        {
                            return '<error type="password">' + _('Bad or missing password') + '</error>';
                        }
                        return '<success>true</success>';
                    }
                );

                //  APIRoute '/change/' -> update session, give nonce, etc... change access
                this.addAPIRoute
                (
                    '1.0',
                    /change/,
                    function( connection, matches )
                    {
                        //  session+acl management
                        _manageSessionAndAcl( connection );

                        if( connection.request.method!=='POST' || !connection.session.isValidNonce )
                        {
                            return;
                        }
                        if( connection.isOwner==false )
                        {
                            return '<error type="access">' + _('Change method only available to the application owner') + '</error>';
                        }

                        //  _uniteActionListener_change -> an error object ?
                        if( _uniteActionListener_change( connection ) )
                        {
                            return '<error type="password">' + _('Bad or missing password') + '</error>';
                        }
                        return '<success><magic>'+ _accessTypes.limited.meta.magic +'</magic></success>';
                    }
                );
            }
        },
        'initialize'
    );
}