/**
 * @fileOverview A Help plugin for Yusef.
 * @author <a href="mailto:antonio.afonso@opera.com">António Afonso</a>
 * @version 0.1
 *
 * Methods:
 * - Yusef.plugins.help.addHelpSection(id, name, content)
 *   Adds a section to be displayed on the /help section, the format used can be found in /templates/help.html. The <code>content</code> parameter may include HTML tags.
 * 
 * - Yusef.plugins.help.addHelpHtml(html)
 *   Adds a piece of HTML to the /help section.
 *
 * API:
 * - /api/help/1.0/section/&lt;section name&gt;
 *   Returns a JSON structure (inside the &lt;result&gt; tag) with the HTML code representing a specific section. Section names are the id of the section to retrieve.
 */
(function()
{
    var self = this;
    var _path;
    var _pluginPath = 'yusef_plugins_help/';
    var _data = {
        sections: [],
        html    : []
    };
    
    /**
     * @plugin help
     *
     * @method addHelpSection
     * @param id
     * @param name
     * @param section
     *
     * @method addHelpHtml
     * @param @html
     */
    Yusef.addPlugin
    (
        'help',
        {
            addHelpSection  : addHelpSection,
            addHelpHtml     : addHelpHtml,
            getData         : getData
        },
        function(){init(this);}
    );
    
    function init(plugin)
    {
        _path = plugin.path;
        
        var publicHtml  = Yusef.mountpoints.application.resolve( _path + '/public_html' );
        Yusef.shareStaticDirectory( publicHtml, _pluginPath );
        
        Yusef.addSectionListener('help', handleHelp, {ui: true});
        plugin.addAPIRoute( '1.0', /section\/(.*)/, handleAPISection );
        
        addGlobalHelpSection();
    }
    
    function getData(connection)
    {
        return {};
    }
    
    /***********
     * Section *
     ***********/
    function handleHelp(connection)
    {
        var template = getTemplate(connection);
        node = template.select();
        var helpTabs = [];
        helpTabs.push({"title": opera.io.webserver.currentServiceName, "section": "", "url": opera.io.webserver.currentServicePath });
        if(Yusef.plugins.activityStream)
        {
            helpTabs.push({"title": _("Activity Feed"), "section": "activity"});
        }
        helpTabs.push({"title": _("Help"), "section": "help", "class": 'active'});
        
        return {
            content : node,
            styles  : [{href: _pluginPath+'style.css'}],
            tabs    : Yusef.plugins.ui.generateTabs(helpTabs, connection)
        };
    }

    /*******
     * API *
     *******/
    function handleAPISection(connection, matches)
    {
        var section = '#'+matches[0]+'_collapsed';
        var template = getTemplate(connection);
        var elements = template.select(section);
        
        if( elements && elements.length > 0 )
        {
            //var result = {
            //    html    : elements[0].innerHTML
            //}
            //return '<![CDATA[' + JSON.stringify(result) + ']]>';
            return '<![CDATA[' + elements[0].innerHTML + ']]>';
        }
    }
    
    /*
    Yusef.plugins.ui && Yusef.plugins.ui.extend
    (
        'generateTabs',
        function extendedGenerateTabs(tabs, connection)
        {
            tabs = tabs || [];
            tabs.push({title: 'Help'});
            
            return extendedGenerateTabs.proceed(tabs, connection);
        }
    );
    */
    
    /**
     * Creates a help section to be shown under /help
     * @public
     *
     * @param   {String}    id      The HTML id to identify this section.
     * @param   {String}    name    The name of the section.
     * @param   {String}    content The contents of the section, can include HTML tags.
     */
    function addHelpSection(id, name, content)
    {
        _data.sections.push({id: id, name: name, content: content});
    }
    
    /**
     * Adds HTML to the /help section
     * @public
     *
     * @param   {String}    html    The HTML text to be added.
     */
    function addHelpHtml(html)
    {
        _data.html.push(html);
    }
    
    function addGlobalHelpSection()
    {
        //addHelpSection
        //(
        //    "openSharedFolder",
        //    _("Open the shared folder"),
        //    _("Open the folder with your shared content.")
        //);
        //
        //addHelpSection
        //(
        //    "selectSharedFolder",
        //    _("Select the shared folder"),
        //    _("Select the folder with your shared content.")
        //);
    }
    
    function getTemplate( connection )
    {
        var data = {
            servicePath: opera.io.webserver.currentServicePath.slice(0, -1),
            serviceName: opera.io.webserver.currentServiceName,
            isOwner:     connection.isOwner,
            username:    opera.io.webserver.userName,
            OS:          'win',
            pluginPath:  _pluginPath
        };
        var template = new Markuper(_path + '/templates/main.html', data);
        
        return template.parse(_data);
    }
})();