/**
 * @fileoverview
 * @author <a href="mailto:antonio.afonso@opera.com">António Afonso</a>, Opera Software ASA
 * @author <a href="mailto:davidh@opera.com">David Håsäther</a>, Opera Software ASA
 * @version 0.1
 */

var TRANSLATOR_USE_PO_FILES = false;
/**
 * @class
 *
 * This class is a wrapper around the GetText library to make it easier to use in JavaScript and in strings.
 * <p>There is a specific syntax, which is <a href="http://docs.djangoproject.com/en/dev/topics/i18n/">based on django</a>, to use it in strings. Here are some examples of the syntax:</p>
 * <h2>Simple Text</h2>
 * <pre><code>
 *   <b>{%trans%}</b>This text will be translated<b>{%endtrans%}</b>
 * </code></pre>
 * Everything between <code>{%trans%}</code> and <code>{%endtrans%}</code> will be translated, after parsing, and these tags removed.
 *
 * <h2>Text With Variables</h2>
 * <pre><code>
 *   {%trans <b>with "value" as key</b>%}The value of key is: <b>{key}</b>{%endtrans%}
 *   {%trans with "António" as firstName <b>and "Afonso" as lastName</b>%}Name: {firstName} {lastName}{%endtrans%}
 * </code></pre>
 * <code>{key}</code>, <code>{firstName}</code> and <code>{lastName}</code> will be replaced by <code>value</code>, <code>António</code> and <code>Afonso</code>, respectively, after translation.
 *
 * <h2>Plurals</h2>
 * <pre><code>
 *   {%trans with <b>"3" as counter</b>%}
 *     There is {counter} apple left.
 *   <b>{%plural%}</b>
 *     There are {counter} apples left.
 *   {%endtrans%}
 * </code></pre>
 * Either the first or the second string will be choosen to be used on the translation, depending on the <code>counter</code> parameter and how plurals are defined in the .po file.
 *
 * <h2>Generic Syntax</h2>
 * <pre><code>
 * {%trans [with "&lt;value&gt;" as &lt;key&gt; [and "&lt;value&gt;" as &lt;key&gt;]*]?%}
 *   &lt;text to be translated that references keys in the form of {&lt;key&gt;}&gt;
 *[{%plural%}
 *   &lt;text to be translated that references keys in the form of {&lt;key&gt;}&gt;
 * {%endtrans%}
 * </code></pre>
 *
 * <p>To use the library in JavaScript a call to the {@link #underscore} is needed:</p>
 * <h2>Simple Text</h2>
 * <pre><code>
 * new Translator().underscore( "This text will be translated" );
 * </code></pre>
 *
 * <h2>Text With Variables</h2>
 * <pre><code>
 * new Translator().underscore( "The value of key is: {key}", {key: "value"} );
 * </code></pre>
 *
 * <h2>Plurals</h2>
 * <pre><code>
 * new Translator().underscore( "There is {counter} apple left.", "There are {counter} apples left.", {counter: 3} );
 * </code></pre>
 *
 * See {@link #translate} for more information about translating on Strings and {@link #underscore} for more information about translating in JavaScript.
 */
function Translator( _config )
{
    var self        = this;
    var _config     = _config || {domain : 'default'};
    var _wl         = new WidgetLocalization();
    var _fileName   = _config.domain+'.json';
    var _poFileName = _config.domain+'.po';
    var _cache      = {};
    var _langs;
    var _gt;

    function init()
    {
        self.setLanguages( _config.langs );
    }

    /**
     * Finds the most appropriate file for translation and returns it as an Object.
     * Tries to find a translation file based on the order defined in the <code>langs</code> Array, if none is found then the default one will be used.
     *
     * @see #.getDefaultLocaleData for more information about the default locale.
     * @private
     */
    function getLocaleData( langs )
    {
        langs = langs || _langs;
        var file;
        var path;
        _wl.findLocaleFiles( langs, _poFileName );

        if( !langs || langs.split(',')[0] in {'en':true, 'en-US':true} )
        {
            locale_data = self.getDefaultLocaleData();
        }
        // If there's a .po file, use it instead, meant for testing only
        else if( TRANSLATOR_USE_PO_FILES && _poFileName in _wl.files )
        {
            path = _wl.files[_poFileName];
            locale_data = {service: Gettext.prototype.parse_po( readFile(path) ) };
        }
        else
        {
            _wl.findLocaleFiles( langs, _fileName );
            path = _wl.files[_fileName];
            locale_data = readFile( path );
            if( locale_data ) { locale_data = JSON.parse( locale_data ); }
        }

        // no file was found, use the defaults
        if( locale_data == null )
        {
            locale_data = self.getDefaultLocaleData();
        }

        return locale_data;
    }

    /**
     * Returns the default locale data when there's none, instead of translating it will just return the key as the value of translation.
     *
     * @returns {Object}    The Dictionary to be used as a stub.
     */
    this.getDefaultLocaleData = function()
    {
        var locale_data = {};

        locale_data[_config.domain] =
        {
            '':
            {
                'plural-forms'  : "nplurals=2; plural=(n != 1);"
            }
        };

        return locale_data;
    }

    /**
     * Helper function to read a file.
     *
     * This function uses memoization, meaning that it will cache the results for a given input and will always return the same results for that input.
     *
     * @param   {String}    path    The path to the file. It can also be a HTTP(S) URI.
     *
     * @private
     */
    function readFile( path )
    {
        if( _cache[path] ) { return _cache[path]; }
        
        // if remote
        if( path && path.toLowerCase().search(/https?:\/\//) == 0 )
        {
            _cache[path] = _gt.sjax(path);
        }
        
        // if local
        try
        {
            file = opera.io.filesystem.mountSystemDirectory( 'application' ).resolve( path );
        }
        catch ( ex )
        {
            throw "Error accessing the filsystem: " + ex;
        }

        if( file.exists )
        {
            try
            {
                var stream = file.open( file, opera.io.filemode.READ );
                _cache[path] = stream.read( stream.bytesAvailable, 'UTF-8' );
            }
            catch( ex )
            {
                throw 'Error reading file "' + path + '": ' + ex;
            }
        }

        return _cache[path];
    }

    /**
     * Parses the variables' declaration part of a <code>{%trans%}</code> block.
     *
     * @param   {String}    string  The String to parse.
     * @returns {Object}            A Dictionary like object with the name of the variables as keys.
     *
     * @private
     */
    function getVariables( string )
    {
        var variables = {};
    
        var regexp = /"((?:\\\\|\\"|[^"])+)"\s*as\s*(\w+)/g;
        string.replace( regexp, function( str, value, name )
        {
            variables[name] = value.replace(/\\(.)/g, '$1');
        });
    
        return variables;
    }

    /**
     * Parses all Strings wrapped in <code>{}</code> and substitute them with the values found in the <code>variables</code> object by indexing it using the String inside the <code>{}</code>.
     *
     * @param   {String}    string      The String with <code>{key}</code> syntax.
     * @param   {Object}    variables   The Dictionary-like Object to be used on replace.
     * @returns {String}                The String after parsing and replacing.
     *
     * @private
     */
    function replaceVariables( string, variables )
    {
        return string.replace( /\{(.*?)\}/g, function( str, g1 )
        {
            return variables[g1];
        });
    }

    /**
     * Finds all <code>{%trans%}</code> blocks in <code>string</code> and parses them.
     *
     * @param   {String}    string      The string to be parsed.
     * @param   {Array}     [langs]     The languages array.
     * @param   {Function}  [callback]  The callback function to be called when a <code>{%trans%}</code> block is found, The list of parameters is <code>callback( <i>singularId</i>, <i>pluralId</i>, <i>variables</i> )</code>.
     * @returns {String}                The <code>string</code> parsed and translated according to <code>langs</code>.
     *
     * @see #setLanguages for more information about the languages array.
     */
    this.translate = function( string, langs, callback )
    {
        var blockRegexp = /\{%\s*trans\s*(?:with)?([\S\s]*?)\s*%\}([\S\s]*?)\{%\s*endtrans\s*%\}/g;
        callback = callback || self.underscore;

        string = string.replace( blockRegexp, function(str, g1, g2)
        {
            var n;
            var lineNr = string.substring(0, blockRegexp.lastIndex-str.length).split(/\n/g).length;
            var phrases = g2.split( /\{%\s*plural\s*%\}/ );
            var vars = getVariables( g1 );

            //for (var i = 0, phrase; phrase = phrases[i]; i++)
            //{
            //    phrases[i] = phrase.replace(/^\s*|\s*$/g, '');
            //}

            //n = parseInt(vars['counter'], 10);
            //var index = (isNaN(n) || n == 1)?0:1;
            //return replaceVariables( gt.gettext(phrases[index]), vars );
            return callback( phrases[0], phrases[1], vars, lineNr, str );
        });

        return string;
    }

    /**
     * Sets the languages that can be used for translation.
     *
     * @param   {Array} languages   The array of language codes to be used in order of preference. The language code must be the same as defined by <a href="http://www.ietf.org/rfc/rfc1766.txt">RFC 1766</a>.
     */
    this.setLanguages = function( languages )
    {
        _langs = languages;

        var locale_data = getLocaleData( _langs );
        // HACK: Workaround for the Gettext static _locale_data property.
        // If an empty locale_data is supplied then the previous loaded
        // locale_data will prevail and will be used instead of the new empty
        // one.
        Gettext._locale_data = undefined;
        _gt = new Gettext( { domain: _config.domain, locale_data: locale_data } );
    }

    /**
     * Translate a string based on its singular or plural ID.
     *
     * @param   {String}    singularId  The singular ID for the string to be translated.
     * @param   {String}    [pluralId]  The plural ID for the string to be translated.
     * @param   {Object}    variables   Variables to be used after translation, this object will be used as a dictionary to parse and substitute strings wrapped in <code>{}</code>.
     * @returns {String}                The translated string.
     */
    this.underscore = function(singularId /*, optional */)
    {
        var pluralId;
        var variables;
        var arg1 = arguments[1]; // pluralId or variables
        // TODO: Talk with david about this
        if (typeof arg1 == "object")
        {
            variables = arg1 || {};
        }
        else
        {
            pluralId = arg1;
            variables = arguments[2];
        }

        if (variables && variables.counter && parseInt(variables.counter, 10) !== 1)
        {
            return replaceVariables(_gt.gettext(pluralId), variables);
        }
        else
        {
            return replaceVariables(_gt.gettext(singularId), variables);
        }
    }

    /**
     * Lazy {@link #underscore} function
     * 
     * This function is just like {@link #underscore} except that it delays the actual translation until the code tries to read its value.
     * This delay might be useful when a string needs to be set up at start-up time but the translation itself should only happen later, during a request.
     *
     * @see #underscore
     */
    this.lazyUnderscore = function() {
        var args = arguments;
        return {
            toString: function() {
                return self.underscore.apply(self, args);
            }
        };
    };

    /**
     * Sets the array of languages to use from the <code>Accept-Language</code> HTTP header found in <code>connection.request.headers</code>.
     *
     * @param   {WebServerConnection}   connection The connection object.
     */
    this.setLanguagesFromHttp = function( connection )
    {
        var langs = (connection.request.getRequestHeader( 'Accept-Language' )||[])[0];
        this.setLanguages( langs );
    };

    /**
     * Function to be used an a REST request.
     * It will send a JavaScript script with the necessary code to translate strings on the client using the client's prefered languages.
     *
     * @param   {WebServerConnection}   connection The connection object.
     */
    this.sendClientScript = function( connection )
    {
        self.setLanguagesFromHttp( connection );

        var data = {localeData: JSON.stringify( getLocaleData() )};
        var tmpl = new Markuper
        (
            '/libraries/yusef/plugins/translation/templates/client.js',
            data,
            {type: 'text'}
        );

        connection.response.write( tmpl.parse().text() );
        connection.response.close();
    };

    init();
}
