/**
 * Copyright (c) 2009, Opera Software ASA
 * All rights reserved.
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of Opera Software ASA nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY OPERA SOFTWARE ASA AND CONTRIBUTORS ``AS IS''
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL OPERA SOFTWARE ASA AND CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/**
 * @fileoverview
 * UI Yusef Plugin.
 *
 * @author António Afonso, Opera Software ASA (antonio.afonso@opera.com)
 * @version 0.1
 */
Yusef.addPlugin
(
    'ui',
    /**
     * This plugin is meant to be used as a way to provide a common layout to a service. It adds a decoration around a service's output.
     * <p>The common layout will always be added when the output of the service request listener is a <code>String</code> and the connection is still open.</p>
     * <p>In order to create an exception to this flow and prevent the plugin from adding a common layout the service might just return a <code>String</code> or close the connection.</p>
     * <p>This plugin also has support for skins which means that it is possible to create a different skin than the one provided by default. In order to tell the plugin to use a specifc skin a call to {@link setSkin} is needed.</p>
     * <p>The skins must be stored in {@link #.SKINS_DIR}.</p>
     *
     */
    function Yusef_plugins_ui()
    {
        var self            = this;
        /**
         * @constant Defines the location of the skins
         */
        this.SKINS_DIR      = '/libraries/yusef/plugins/ui/skins/';
        /**
         * @constant URL path where this plugin specific files are exposed
         */
        this.PLUGIN_PATH    = 'yusef_plugins_ui';
        /**
         * @constant Id used to publish content on the {@link PSO} object.
         */
        var ERR             = 'Yusef.plugins.ui';

        var skinName        = 'default';
        var _styleSheet;

        /**
         * Makes all files in <code>{@link #.SKINS_DIR}/&lt;skin&gt;/public_html</code> accessible on the webserver through the  <code>static/{@link #.PLUGIN_PATH}/&lt;skin&gt;</code> URL.
         *
         * @private
         */
        (function shareSkinsPublicFiles()
        {
            var skinsDir = Yusef.mountpoints.application.resolve( self.SKINS_DIR );
            skinsDir.refresh();
            for( var i = 0, skin; skin = skinsDir[i]; i++ )
            {
                if( !skin.isDirectory || skin.name[0] == '.' ) { continue; }
                Yusef.shareStaticDirectory(skin.resolve('public_html'), self.PLUGIN_PATH + '/' + skin.name + '/');
            }
        })();

        /**
         * Returns all the data associated with the UI plugin and the service itself.
         * <p>This data is an abstracion of several different sources such as <code>opera.io.webserver.*</code> and <code>connection.*</code>, and is meant to be used on templates or some other means of generating output to the client.</p>
         * <p>Note: Some of this data will be available under Yusef.getData() in the future.</p>
         *
         * @param   {WebServerConnection}   connection  The connection object from which to extract specific information like, for instance, the request object.
         * @returns {Object}                            Collection of key/value pairs in an object with the data request.
         */
        this.getData = function( connection )
        {
            var data = {};
            var title;

            if (opera.io.webserver.currentServicePath == "_root")
            {
                title = opera.io.webserver.userName + " | " + opera.io.webserver.deviceName;
            }
            else
            {
                title = opera.io.webserver.currentServiceName;
            }
            if( connection.isOwner )
            {
                data.title = _("{userServiceOrDeviceName} | Opera Unite", {
                    userServiceOrDeviceName : title
                });
            } else {
                data.title = _("{userServiceOrDeviceName} by {username} and Opera Unite", {
                    userServiceOrDeviceName : title,
                    username: opera.io.webserver.userName
                });
            }
            

            data.public_html = opera.io.webserver.currentServicePath.slice(0, -1) + '/static/' + self.PLUGIN_PATH + '/' + skinName;
            data.style = _styleSheet;

            return data;
        }

        /**
         * Creates an HTML string using the information stored in the <code>data</code>, it currently uses <code>Markuper</code> as a template engine to create the HTML string.
         * <p>The template used for <code>Markuper</code> is <code><code>{@link #.SKINS_DIR}/{@link skinName}/main.html</code>.</p>
         *
         * @param   {Object}    data    Data to be used in the creation of the HTML string.
         * @returns {String}            The HTML string.
         *
         * @see #setSkin to know how to change the skin used.
         */
        this.formatData = function( data )
        {
            var skinPath = getSkinPath( skinName );

            try
            {
                var tmpl = new Markuper( skinPath + "/main.html", data );
                return tmpl.parse(data.data).html();
            }
            catch( ex )
            {
                return ex.message;
            }
        }

        /**
         * Sets the skin to be used for creating the HTML string. By default it is used the <code>default</code> skin.
         * <p>The skin must reside in {@link #.SKINS_DIR}.</p>
         *
         * @param   {String}    skin    The directory name where the skin is, this name must be relative to {@link #.SKINS_DIR}.
         *
         * @see #formatData
         */
        this.setSkin = function( skin )
        {
            skinName = skin;
        }

        this.setStyleSheet = function( src )
        {
            _styleSheet = src;
        }

        /**
         * Generates tabs that can be used in the skin template
         *
         * @param {Array}               tabs        An array of objects with at least a title property. Other properties available are: 'section' for specifying the section, 'root' for specifying that the tab is for the root section, 'hideFromVisitors' for specifying that this tab should be hidden from visitors. If no section is specified, the section will be same as the title, but lowercase.
         * @param {WebServerConnection} connection  The connection object from which to extract specific information like, for instance, the request object.
         */
        this.generateTabs = function( tabs, connection )
        {
            var servicePath = opera.io.webserver.currentServicePath;
            var url = connection.request.uri;
            if (url.lastIndexOf("/")>1) url = url.slice(0, url.indexOf("/", 1) + 1);
            if (url.slice(-1) == "/" && url != servicePath) { url = url.slice(0, -1); }

            for (var i = 0, tab; tab = tabs[i]; i++)
            {
                if (tab.section === undefined) { tab.section = tab.title.toLowerCase(); }
                if (tab.section !== "") tab.section = tab.section + '/';
                tab.url = servicePath + tab.section;
                if (tab.isRoot) { tab.url = servicePath; }
                //tab.class = (tab.url == url) ? 'active' : '';
                tab.hidden = tab.hideFromVisitors && !connection.isOwner;
            }

            return tabs;
        }

        /**
         * Returns the skin path relative to the <code>application</code> mountpoint.
         *
         * @param   {String}    skinName    The name of the skin.
         * @returns {String}                The path where the skin stands.
         */
        function getSkinPath( skinName )
        {
            return self.SKINS_DIR + skinName;
        }

        Yusef.extend
        (
            'addSectionListener',
            function extendedAddSectionListener( name, listener, options )
            {
                // only add UI if specified unless it's a default Section
                options = options || {};
                var defaultSections = {'content':true, '_index':true};

                var hasUI = (options && 'ui' in options) ? !!options.ui : (name in defaultSections);

                if( hasUI ) { options.ui = true; }

                return extendedAddSectionListener.proceed();
            }
        );

        Yusef.extend
        (
            'addRoute',
            /**
             * Extends {@link yusef#addRoute} in order to modify the registered <code>method</code> to make its output be decorated by a common layout defined in <code>{@link #.SKINS_DIR}/{@link skinName}/</code>.
             * <p>The new method will call the old one and if the returned value is not a <code>String</code>, and if the <p>connection</p> is still open, then the returned object will be used as an argument to the {@link  markuper.Template#parse} <code>Markuper</code> function.</p>
             * <p>When a service includes this plugin to work under a specifc section, the section code should return an object meaningful to the <code>Markuper</code> skin currently in use.</p>
             *
             * @param   {Object}    options Dictionary of options.
             * @param   {Object}    [options.ui]  Whether or not the UI should be used (default is true only for 'content' and '_index').
             *
             * @private
             */
            function extendedAddRoute( pattern, method, options )
            {
                //  modify the listener to inject a common UI
                var modifiedMethod = function( connection )
                {
                    // get the output from the original listener
                    var methodData = method.apply( this, arguments ) || {};

                    if( connection.closed===false && !connection.pending )
                    {
                        var data = Yusef.getData( connection, true );
                        if( typeof(methodData) == 'string'
                         || (typeof(methodData) == 'object') && methodData.constructor == String)
                        {
                             return methodData;
                        }
                        data.data = methodData;
                        output = self.formatData( data );
                        return output;
                    } else {
                        return {};
                    }

                };

                if( options && !options.ui ) { modifiedMethod = method; }

                return extendedAddRoute.proceed( pattern, modifiedMethod, options );
            }
        );

		/**
         * Unite Section listener to error
         * <p>For common error page in all services</p>
         *
         */
        Yusef.extend
		(
            'emitErrorPage',
            function( connection, errorCode )
            {
                var data = Yusef.getData( connection, true );
                data.errorCode  = errorCode;
                var skinPath    = getSkinPath( skinName );
                var tmpl        = new Markuper( skinPath + '/error.html', data );
                tmpl.parse( data );

                data.data =
                {
                    title   : 'Error',
                    isError : true,
                    content : tmpl.select( '#error' ),
                    styles  : [{href: 'style.css'}],
                    tabs: [{title: opera.io.webserver.currentServiceName, url: opera.io.webserver.currentServicePath}]
                };
                
                connection.response.setStatusCode( '404', 'Not Found' );
                var output =  self.formatData( data );
                
                if( Yusef.plugins.translation )
                {
                    output = Yusef.plugins.translation.translate(connection, output);
                }
                
                return output;
            }
        );
    }
);
